<?php

namespace App\Http\Controllers\Api;

use App\ServiceLog;
use App\UserMeta;

use App\Http\Controllers\Controller;
use App\Http\Requests\StoreServiceLogRequest;
use App\Transformers\ServiceLogTransformer;
use App\Transformers\ServiceLogCsvTransformer;
use League\Fractal\Pagination\IlluminatePaginatorAdapter;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Carbon\Carbon;

class ServiceLogController extends Controller
{
    /**
     * Display a listing of the resource.
     *
     * @return \Illuminate\Http\Response
     */
    public function index(Request $request)
    {
        $events = ServiceLog::select(
                'service_logs.id',
                'service_logs.created_at',
                'service_logs.user_id',
                'service_logs.opened_by',
                'service_logs.closed_by',
                'service_logs.start',
                'service_logs.stop',
                'service_logs.scheduled_date',
                'service_logs.status',
                'service_logs.note',
                'service_logs.service_type_id',
                'service_logs.task_id',
                'service_logs.tail_number_id',
                'service_logs.station_id',
                'service_logs.airline_id',
                'service_logs.gate_id',
                'service_logs.geo_id',
                DB::raw('TIME_TO_SEC(service_logs.stop) - TIME_TO_SEC(service_logs.start) AS duration'),
                DB::raw('UNIX_TIMESTAMP(service_logs.start) - UNIX_TIMESTAMP(service_logs.created_at) AS turnaround')
            )
            ->join('tasks', 'service_logs.task_id', '=', 'tasks.id')
            ->join('fleet', 'service_logs.tail_number_id', '=', 'fleet.id')
            ->join('barcodes', 'service_logs.tail_number_id', '=', 'barcodes.fleet_id')
            ->join('gates', 'service_logs.gate_id', '=', 'gates.id')
            ->join('user_meta', 'service_logs.user_id', '=', 'user_meta.id')
            ->join('geo', 'service_logs.geo_id', '=', 'geo.id', 'left')
            ->where('service_logs.service_type_id', 12)
            ->where('stop', '!=', '0000-00-00 00:00:00')
            ->whereNotNull('stop')
            ->with(['fleet' => function ($query) {
                $query->withTrashed();
            }, 'gate', 'station', 'fleet.aircraftType', 'task', 'userMeta', 'fleet.barcodeHistory', 'openedBy', 'closedBy'])->filter($request)
            ->order($request);

        if (request()->segment(5) == 'csv') {
            $eventsCollection = $events->get();

            return fractal()
                ->collection($eventsCollection)
                ->transformWith(new ServiceLogCsvTransformer)
                ->toArray();
        } else {
            $events = $events->paginate(15);
            $eventsCollection = $events->getCollection();

            return fractal()
                ->collection($eventsCollection)
                ->transformWith(new ServiceLogTransformer)
                ->paginateWith(new IlluminatePaginatorAdapter($events))
                ->toArray();
        }
    }

    public function stats(Request $request)
    {
        $events = ServiceLog::select(DB::raw('TIME_TO_SEC(service_logs.stop) - TIME_TO_SEC(service_logs.start) AS duration'))
            ->where('service_type_id', 12)
            ->where('stop', '!=', '0000-00-00 00:00:00')
            ->whereNotNull('stop')
            ->with('fleet')
            ->with('gate')
            ->filter($request)
            ->get();

        if ($events->count() > 0) {
            foreach ($events as $record) {
                $duration = strtotime('1970-01-01 ' . gmdate('H:i:s', $record->duration) . 'GMT');
                $durations[] = $duration;
            }

            sort($durations);

            $stats['low']   = gmdate('H:i:s', ($durations[0]));
            $stats['high']  = gmdate('H:i:s', ($durations[$events->count() - 1]));
            $stats['average'] = gmdate('H:i:s', (array_sum($durations) / $events->count()));

            return $stats;
        }
    }

    public function open()
    {
        $user = \App\User::findOrFail(Auth::user()->id);
        $userLegacy = \App\UserLegacy::find($user->legacy_id);
        $userGroupId = $userLegacy->group_id;
        $userTaskPermissions = UserMeta::taskPermissions();

        if (is_array($userTaskPermissions)) {
            $events = ServiceLog::where('service_type_id', 12)
                ->where(function ($query) {
                    $query->where('stop', null)
                            ->orWhere('stop', '0000-00-00 00:00:00');
                })
                ->where(function ($query) {
                    $query->where('status', 0)
                            ->orWhere('status', 2);
                })
                ->whereIn('task_id', $userTaskPermissions)
                ->get();
        } elseif ($userGroupId == 6) {
            $events = ServiceLog::where('service_type_id', 12)
                ->where(function ($query) {
                    $query->where('stop', null)
                            ->orWhere('stop', '0000-00-00 00:00:00');
                })
                ->where(function ($query) {
                    $query->where('status', 2);
                })
                ->get();
        } else {
            $events = ServiceLog::where('service_type_id', 12)
                ->where(function ($query) {
                    $query->where('stop', null)
                            ->orWhere('stop', '0000-00-00 00:00:00');
                })
                ->where(function ($query) {
                    $query->where('status', 0)
                            ->orWhere('status', 2);
                })
                ->get();
        }

        if (count($events) > 0) {
            return fractal()
                ->collection($events)
                ->transformWith(new ServiceLogTransformer)
                ->toArray();
        }
    }

    public function batch($service, $userId)
    {
        $events = ServiceLog::where('service_type_id', 12)
            ->where('task_id', 534)
            ->where('user_id', $userId)
            ->where(function ($query) {
                $query->where('stop', null)
                        ->orWhere('stop', '0000-00-00 00:00:00');
            })
            ->get();

        if (count($events) > 0) {
            return fractal()
                ->collection($events)
                ->transformWith(new ServiceLogTransformer)
                ->toArray();
        }
    }

    public function recentAudits()
    {
        $events = ServiceLog::where('service_type_id', 12)
            ->where('task_id', 534)
            ->where('start', '>=', date('Y-m-d H:i:s', strtotime('-12 hours')))
            ->get()
            ->pluck('tail_number_id');

        if (count($events) > 0) {
            return response()->json(['data' => $events], 200);
        }
    }

    public function requests()
    {
        $events = ServiceLog::where('service_type_id', 12)
            ->where('status', 1)
            ->where(function ($query) {
                $query->where('stop', null)
                        ->orWhere('stop', '0000-00-00 00:00:00');
            })
            ->orderBy('utp_reason', 'desc')
            ->orderBy('scheduled_date', 'asc')
            ->get();

        return fractal()
            ->collection($events)
            ->transformWith(new ServiceLogTransformer)
            ->toArray();
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param  \Illuminate\Http\Request  $request
     * @return \Illuminate\Http\Response
     */
    public function store(StoreServiceLogRequest $request)
    {
        $event = new ServiceLog;
        $event->user_id = $request->user_id; //Auth::user()->id;
        $event->service_type_id = 12;
        $event->station_id = $request->station_id;
        $event->airline_id = $request->airline_id;
        $event->task_id = $request->task_id;
        $event->gate_id = $request->gate_id;
        $event->tail_number_id = $request->tail_number_id;
        $event->scheduled_date = $request->scheduled_date;
        $event->start = date('Y-m-d H:i:s');

        $event->save();

        return response()->json(['data' => $event], 200);
    }

    /**
     * Display the specified resource.
     *
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function show($id)
    {
        $response = ServiceLog::findOrFail($id);
        $mymedia = $response->getMedia('terminal');
        $response->media = $response->getMedia('terminal');

        if (!empty($response->reference_table)) {
            $fractal = fractal($response, new TrainingLogTransformer)
                ->toArray();
        } else {
            $fractal = fractal($response, new ServiceLogTransformer)
                ->toArray();
        }
        return response()->json($fractal, 200);
    }

    /**
     * Update the specified resource in storage.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function update(Request $request, $id)
    {
        $event = ServiceLog::findOrFail($id);

        if (isset($request->note_only) && $request->note_only == 1) {
            $event->note = $request->note;
            $event->save();
            return response()->json(['data' => $event], 200);
        }

        if (isset($request->geo_id)) {
            $event->geo_id = $request->geo_id;
            $event->save();
            return response()->json(['data' => $event], 200);
        }

        if (isset($request->gate_id) && $request->gate_id_only === true) {
            $event->gate_id = $request->gate_id;
            $event->save();
            return response()->json(['data' => $event], 200);
        }

        if (isset($request->gate_id)) {
            $event->gate_id = $request->gate_id;
            $event->start = date('Y-m-d H:i:s');
            if (isset($event->closed_by) && $event->closed_by != null) {
                $event->closed_by = $request->closed_by;
            }
            if ($event->status == 1) {
                $event->status = 2;
            }
            $event->save();
            return response()->json(['data' => $event], 200);
        }

        if (isset($request->cmd) && $request->cmd == 'update') {
            $event->tail_number_id = $request->fleet_id;
            $event->gate_id = $request->gate_id;
            $event->start = $request->start;
        } else { // stop service
            if ($event->status < 1) {
                $event->note = $request->note;
            }
            if (isset($event->closed_by) && $event->closed_by != null) {
                $event->closed_by = $request->closed_by;
            }
        }

        if ($request->cmd == 'complete') {
            $event->stop = date('Y-m-d H:i:s');
        }

        // Set service request info
        if ($event->status == 2) {
            $event->status = 3;
            $event->note = 'REQUEST: ' . $event->note . "\r" . 'SERVICE: ' . $request->note;
            $event->closed_by = $request->closed_by;
            $event->stop = date('Y-m-d H:i:s');
        } elseif ($event->status == 1) {
            $event->status = 2;
        }
        //dd($event);
        $event->save();

        return response()->json(['data' => $event], 200);
    }

    public function setSchedule($entry)
    {
        $schedule = \App\Schedule::where([
            'tail_number_id' => $entry['fleet']['data']['id'],
            'task_id' => $entry['task']['data']['id'],
            //'date' => $entry['scheduled_date'],
        ])->first();

        $proposed = new Carbon('first day of next month');
        $schedule->date = $proposed->format('Y-m-d');
        //dd($schedule);
        $schedule->save();
    }

    public function stopBatch($service)
    {
        // need user legacy id
        $user = \App\User::findOrFail(Auth::user()->id);

        $batch = $this->batch($service, $user->legacy_id);

        if (isset($batch['data'])) {
            // start date is the first entry.
            $first = collect($batch['data'][0]);

            foreach($batch['data'] as $entry)
            {
                $existing = ServiceLog::findOrFail($entry['id']);
                $existing->start = $first['start'];
                $existing->stop = date('Y-m-d H:i:s');
                $existing->save();

                $this->setSchedule($entry);
            }
            return response()->json(['data' => ['batch' => 'success']], 200);
        }
        return response()->json(['data' => ['batch' => 'failed']], 200);
    }

    public function upload(Request $request, $id)
    {
        $event = ServiceLog::findOrFail($id);
        $event->addMedia($request->file)
            ->toMediaCollection('terminal');

        return response()->json(['data' => ['upload' => 'success']], 200);
    }

    /**
     * Remove the specified resource from storage.
     *
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function destroy($id)
    {
        $event = ServiceLog::findOrFail($id);
        $event->delete();
    }
}
