<?php

namespace App\Http\Controllers;

use App\ServiceLog;
use Carbon\Carbon;

use App\Http\Controllers\Controller;
use App\Http\Requests\StoreServiceLogRequest;
use App\Transformers\ServiceLogTransformer;
use App\Transformers\ServiceLogCsvTransformer;
use League\Fractal\Pagination\IlluminatePaginatorAdapter;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;

class ServiceLogController extends Controller
{
    /**
     * Display a listing of the resource.
     *
     * @return \Illuminate\Http\Response
     */
    public function index(Request $request)
    {
        $events = ServiceLog::select(
                'service_logs.id',
                'service_logs.created_at',
                'service_logs.user_id',
                'service_logs.opened_by',
                'service_logs.closed_by',
                'service_logs.start',
                'service_logs.stop',
                'service_logs.scheduled_date',
                'service_logs.status',
                'service_logs.note',
                'service_logs.service_type_id',
                'service_logs.task_id',
                'service_logs.tail_number_id',
                'service_logs.station_id',
                'service_logs.airline_id',
                'service_logs.gate_id',
                'service_logs.geo_id',
                DB::raw('TIME_TO_SEC(service_logs.stop) - TIME_TO_SEC(service_logs.start) AS duration'),
                DB::raw('UNIX_TIMESTAMP(service_logs.start) - UNIX_TIMESTAMP(service_logs.created_at) AS turnaround')
            )
            ->join('tasks', 'service_logs.task_id', '=', 'tasks.id')
            ->join('fleet', 'service_logs.tail_number_id', '=', 'fleet.id')
            ->join('barcodes', 'service_logs.tail_number_id', '=', 'barcodes.fleet_id')
            ->join('gates', 'service_logs.gate_id', '=', 'gates.id')
            ->join('user_meta', 'service_logs.user_id', '=', 'user_meta.id')
            ->join('geo', 'service_logs.geo_id', '=', 'geo.id', 'left')
            ->where('service_logs.service_type_id', 12)
            ->where('stop', '!=', '0000-00-00 00:00:00')
            ->whereNotNull('stop')
            ->with(['fleet' => function ($query) {
                $query->withTrashed();
            }, 'gate', 'station', 'fleet.aircraftType', 'task', 'userMeta', 'fleet.barcodeHistory', 'openedBy', 'closedBy'])->filter($request)
            ->order($request);

        if (request()->segment(4) == 'csv') {
            $eventsCollection = $events->get();

            return fractal()
                ->collection($eventsCollection)
                ->transformWith(new ServiceLogCsvTransformer)
                ->toArray();
        } else {
            $events = $events->paginate(15);
            $eventsCollection = $events->getCollection();

            return fractal()
                ->collection($eventsCollection)
                ->transformWith(new ServiceLogTransformer)
                ->paginateWith(new IlluminatePaginatorAdapter($events))
                ->toArray();
        }
    }

    public function stats(Request $request)
    {
        $events = ServiceLog::select(DB::raw('TIME_TO_SEC(service_logs.stop) - TIME_TO_SEC(service_logs.start) AS duration'))
            ->where('service_type_id', 12)
            ->whereNotNull('stop')
            ->with('fleet')
            ->filter($request)
            ->get();

        if ($events->count() > 0) {
            foreach ($events as $record) {
                $duration = strtotime('1970-01-01 ' . gmdate('H:i:s', $record->duration) . 'GMT');
                $durations[] = $duration;
            }

            sort($durations);

            $stats['low']   = gmdate('H:i:s', ($durations[0]));
            $stats['high']  = gmdate('H:i:s', ($durations[$events->count() - 1]));
            $stats['average'] = gmdate('H:i:s', (array_sum($durations) / $events->count()));

            print_r($stats);
        }
    }

    public function open()
    {
        $events = ServiceLog::where('service_type_id', 12)
            ->where('stop', null)
            ->get();

        return fractal()
            ->collection($events)
            ->transformWith(new ServiceLogTransformer)
            ->toArray();
    }

    public function batch()
    {
        $events = ServiceLog::where('service_type_id', 12)
            ->where('stop', null)
            ->where('task_id', 534)
            ->get();

        return fractal()
            ->collection($events)
            ->transformWith(new ServiceLogTransformer)
            ->toArray();
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param  \Illuminate\Http\Request  $request
     * @return \Illuminate\Http\Response
     */
    public function store(StoreServiceLogRequest $request)
    {
        $event = new ServiceLog;
        $event->user_id = Auth::user()->id;
        $event->service_type_id = 12;
        $event->station_id = $request->station_id;
        $event->airline_id = $request->airline_id;
        $event->task_id = $request->task_id;
        $event->gate_id = $request->gate_id;
        $event->tail_number_id = $request->tail_number_id;
        $event->scheduled_date = $request->scheduled_date;
        $event->start = date('Y-m-d H:i:s');

        $event->save();

        return response()->json(['data' => $event], 200);
    }

    /**
     * Display the specified resource.
     *
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function show($id)
    {
        $response = ServiceLog::findOrFail($id);
        $mymedia = $response->getMedia('terminal');
        $response->media = $response->getMedia('terminal');

        /*
        $fractal = fractal()
            ->item($response)
            ->transformWith(new ServiceLogTransformer)
            ->toArray();
        */

        if (!empty($response->reference_table)) {
            $fractal = fractal($response, new TrainingLogTransformer)
                ->toArray();
        } else {
            $fractal = fractal($response, new ServiceLogTransformer)
                ->toArray();
        }


        //dd($fractal);

        return response()->json($fractal, 200);
    }

    /**
     * Update the specified resource in storage.
     *
     * @param  \Illuminate\Http\Request  $request
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function update(Request $request, $id)
    {
        $event = ServiceLog::findOrFail($id);
        $event->stop = date('Y-m-d H:i:s');
        $event->note = $request->note;

        $event->save();

        return response()->json(['data' => $event], 200);
    }

    public function upload(Request $request, $id)
    {
        $event = ServiceLog::findOrFail($id);
        $event->addMedia($request->file)
            ->toMediaCollection('terminal')
            ->useDisk('s3');
    }

    /**
     * Remove the specified resource from storage.
     *
     * @param  int  $id
     * @return \Illuminate\Http\Response
     */
    public function destroy($id)
    {
        //
    }

    /*
    public function fixDates()
    {
        $events = ServiceLog::whereColumn('created_at', '<>', 'start_c')
            ->whereColumn('created_at', '<>', 'updated_at')
            ->whereDate('created_at', '>=', '2018-11-16')
            ->whereDate('created_at', '<', '2018-11-20')
            ->where('service_type_id', '<>', 8)
            ->where('status', '<>', 3)
            ->has('station')
            ->get();

        echo $events->count();

        foreach ($events as $event) {
            if (substr($event->start_c, -3, 3) == ':00' && substr($event->stop_c, -3, 3) == ':00') {
                echo '<p>MANUAL ENTRY: ' . $event->start_c . ' ' . $event->stop_c . '<code></code></p>';
            } else {
                $start = Carbon::parse($event->start_c);
                $stop = Carbon::parse($event->stop_c);
                $diff = $start->diffInMinutes($stop);
                $stop_c = Carbon::parse($event->start_c);
                $stop_c = $stop_c->addMinutes($diff);

                echo '<p>UPDATE ENTRY: ' . $event->id . ' ' . $event->station->timezone_id . ' ' . $event->start_c . ' ' . $event->stop_c . ' ' . $stop_c . '<br/>';
                echo 'DIFF: ' . $diff . '<br/>';

                $new_start = Carbon::parse($event->created_at);
                $new_start->addMinutes($event->station->timezone_offset);

                $new_stop = Carbon::parse($event->created_at);
                $new_stop->addMinutes($event->station->timezone_offset)->addMinutes($diff);

                echo 'UPDATEED TO: &nbsp;&nbsp;' . $event->id . ' ' . $event->station->timezone_id . ' ' . $new_start . ' ' . $new_stop . '</p>';
                // get the timezone
                //print_r($event->created_at);

                $dates = [
                    'start_c' => $event->created_at,
                    'stop_c' => $event->updated_at,
                    'start' => $new_start,
                    'stop' => $new_stop,
                ];
                //print_r($dates);

                $update = ServiceLog::findOrFail($event->id);
                $update->start_c = $event->created_at->toDateTimeString();
                $update->stop_c = $stop_c->toDateTimeString();
                $update->start = $new_start->toDateTimeString();
                $update->stop = $new_stop->toDateTimeString();
                $update->timestamps = false;
                $update->save($dates);
            }
        }
    }
    */
}
