<?php

namespace App;

use App\ServiceType;
use App\Task;
use App\Filters\Schedule\ScheduleFilters;
use App\Filters\Filters;

use Illuminate\Database\Eloquent\SoftDeletes;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Http\Request;
use OwenIt\Auditing\Contracts\Auditable;

class Schedule extends Model implements Auditable
{
    use \OwenIt\Auditing\Auditable;
    use SoftDeletes;

    protected $guarded = [
        'id',
        'created_at',
        'updated_at',
    ];

    protected $hasThese = [
        'fleet',
        'serviceType',
    ];

    public static function withFleet(Request $request)
    {
        return Schedule::select(['schedules.*', 'aircraft_types.title as aircraft'])
            ->where('schedules.deleted', 0)
            ->where('service_type_id', 12)
            ->where('fleet.tail_number', '<', 90000)
            ->join('fleet', 'schedules.tail_number_id', '=', 'fleet.id')
            ->join('gates', 'gates.id', '=', 'fleet.gate_id', 'left')
            ->join('aircraft_types', 'aircraft_types.id', '=', 'fleet.aircraft_type_id')
            ->join('barcodes', 'barcodes.fleet_id', '=', 'fleet.id')
            ->join('airlines', 'airlines.id', '=', 'fleet.airline_id')
            ->when(isset($request['terminal']), function ($q) use ($request) {
                return $q->where('gates.terminal', $request['terminal']);
            });
    }

    public static function compliance(Request $request)
    {
        return Schedule::select([
                'schedules.*',
                'aircraft_types.title as aircraft',
                'clean.date as clean_date',
                'clean.id as clean_id',
                'clean_todo.id as clean_todo',
                'mechanical.date as mechanical_date',
                'mechanical.id as mechanical_id',
                'mechanical_todo.id as mechanical_todo',
                'electrical.date as electrical_date',
                'electrical.id as electrical_id',
                'electrical_todo.id as electrical_todo',
            ])
            ->where('schedules.deleted', 0)
            ->where('schedules.service_type_id', 12)
            ->where('schedules.task_id', 534)
            ->where('fleet.tail_number', '<', 90000)
            ->join('schedules as clean', 'schedules.tail_number_id', '=', 'clean.tail_number_id')
            ->leftJoin('todo as clean_todo', 'clean.id', '=', 'clean_todo.schedule_id')
            ->where('clean.task_id', 531)
            ->where('clean.deleted', 0)
            ->join('schedules as mechanical', 'schedules.tail_number_id', '=', 'mechanical.tail_number_id')
            ->leftJoin('todo as mechanical_todo', 'mechanical.id', '=', 'mechanical_todo.schedule_id')
            ->where('mechanical.task_id', 532)
            ->where('mechanical.deleted', 0)
            ->join('schedules as electrical', 'schedules.tail_number_id', '=', 'electrical.tail_number_id')
            ->leftJoin('todo as electrical_todo', 'electrical.id', '=', 'electrical_todo.schedule_id')
            ->where('electrical.task_id', 533)
            ->where('electrical.deleted', 0)
            ->join('fleet', 'schedules.tail_number_id', '=', 'fleet.id')
            ->join('gates', 'gates.id', '=', 'fleet.gate_id', 'left')
            ->join('aircraft_types', 'aircraft_types.id', '=', 'fleet.aircraft_type_id')
            ->join('barcodes', 'barcodes.fleet_id', '=', 'fleet.id')
            ->join('airlines', 'airlines.id', '=', 'fleet.airline_id')
            ->when(isset($request['terminal']), function ($q) use ($request) {
                return $q->where('gates.terminal', $request['terminal']);
            })
            ->orderBy('barcodes.barcode');
    }

    public function task()
    {
        return $this->belongsTo('App\Task');
    }

    public function fleet()
    {
        return $this->belongsTo('App\Fleet', 'tail_number_id', 'id');
    }

    public function serviceType()
    {
        return $this->belongsTo('App\ServiceType');
    }

    public function todo()
    {
        return $this->hasOne('App\Todo');
    }

    public function barcode()
    {
        return $this->hasOne('App\Barcode', 'fleet_id', 'tail_number_id');
    }

    public function scopeOrder(Builder $builder, Request $request)
    {
        if (isset($request['order']) && $request['order'] == 'tail_number') {
            return $builder->orderIdentifier($request);
        } elseif (!isset($request['order']) || !isset($request['dir'])) {
            return $builder->orderBy('date', 'desc');
        } else {
            return $builder->orderBy($request->order, $request->dir);
        }
    }

    public function scopeOrderIdentifier(Builder $builder, Request $request)
    {
        return $builder->orderBy('fleet.tail_number', array_get($request, 'dir', 'asc'));
    }

    public function scopeFilter(Builder $builder, Request $request, array $filters = [])
    {
        return (new ScheduleFilters($request))->add($filters)->filter($builder->hasThese());
    }

    /*
     * Make sure these are not deleted.
     */
    public function scopeHasThese(Builder $builder, $has = [])
    {
        $has = array_merge($has, $this->hasThese);
        foreach ($has as $it) {
            $builder->has($it);
        }
        return $builder;
    }

    public function scopeTotalScheduled(Builder $builder, int $airline_id, int $service_type_id, ?array $fleet_ids)
    {
        return $builder->selectRaw('schedules.task_id, COUNT(schedules.id) AS total, CONCAT(task_groups.title, ": ", tasks.title) AS service')
            ->join('fleet', 'fleet.id', '=', 'schedules.tail_number_id')
            ->join('tasks', 'tasks.id',  '=',  'schedules.task_id')
            ->join('task_groups', 'task_groups.id','=', 'tasks.group_id', 'left')
            ->when($fleet_ids !== null, function($query) use($fleet_ids) {
                return $query->whereIn('fleet.aircraft_type_id', $fleet_ids);
            })
            ->where('fleet.airline_id', $airline_id)
            ->where('schedules.service_type_id', $service_type_id)
            ->where('schedules.date', '>', '2010-10-10')
            ->where('tasks.deleted', 0)
            ->where('fleet.deleted', 0)
            ->groupBy(['schedules.task_id', 'service']);
    }

    /*
    public function scopePhoDelete(Builder $builder)
    {
        return $builder->where('deleted', 0);
    }

    public function scopeByServiceType(Builder $builder, $id)
    {
        if (!is_numeric($id)) {
            return ServiceType::where('abbr', $id)
                ->schedules();
        } else {
            return ServiceType::find($id)
                ->schedules();
        }
    }

    public function scopeByTask($query, $id)
    {
        return Task::find($id)
            ->schedules();
    }

    public function scopeByAirline($query, $slug)
    {
        return ServiceType::where('abbr', $slug)
            ->first()
            ->schedules();
    }
    */
}
